/* Copyright (C) Eliot Stock - All Rights Reserved
 * Unauthorized copying of this file, via any medium is strictly prohibited.
 * Proprietary and confidential.
 * Written by Eliot Stock, 1@eliotstock.com
 */
#include <stdio.h>

#include "app_util_platform.h"
#include "nrf_drv_gpiote.h"
#include "nrf_gpio.h"
#include "nrf_delay.h"
#include "app_twi.h"

#include "config.h"
#include "board.h"
#include "error.h"
#include "acc_kionix_ktxj2.h"
#include "util.h"
#include "debug_leds.h"

// Driver states
typedef enum
{
    STATE_NOT_INITED,
    STATE_INITED,
    STATE_PROGRAMMED
} kionix_state_t;

static kionix_state_t m_kionix_state = STATE_NOT_INITED;

// TWI instance. ID must differ from that used anywhere else.
static app_twi_t m_twi_kionix = APP_TWI_INSTANCE(0);

// From main.c
volatile extern bool m_moving;

#ifdef BOARD_V05

// Called on motion, when the accelerometer sends us an interrupt.
void on_acc_interrupt_pin_high(nrf_drv_gpiote_pin_t pin, nrf_gpiote_polarity_t action)
{
	m_moving = true;
	debug_led_on(PIN_LED_MOTION_FOR_TRACKING);
}

// Set up the TWI (I2C) interface to the accelerometer. This is repeatable and does nothing the second time it's called.
void acc_kionix_init(void)
{
    uint32_t err_code = 0;

    if (m_kionix_state >= STATE_INITED)
    {
    	printf("Already init'ed. Nothing to do.\n");
    	return;
    }

    // 100 kHz is "normal mode", 400 kHz is "fast mode" in in the I2C standard. 400 kHz is a period of 2.5 us.
	// Heinous SDK gotcha: these values belong in nrf_drv_config.h *as well* as here!
    nrf_drv_twi_config_t const config = {
       .scl                = PIN_KX_SCL,
       .sda                = PIN_KX_SDA,
       .frequency          = NRF_TWI_FREQ_400K,
       .interrupt_priority = APP_IRQ_PRIORITY_LOW
    };

    APP_TWI_INIT(&m_twi_kionix, &config, KIONIX_MAX_PENDING_TRANSACTIONS, err_code);
    APP_ERROR_CHECK(err_code);

	// TODO: Consider removing this test once the driver becomes reliable. We'll know if we're good on acc_kionix_program() anyway.
	// Check we can actually read a register. The WHO_AM_I register is never zero.
	// Note that the variables here must be static. While it looks awfully tempting to move all this stuff into a function or a macro,
	// don't. It just complicates things. See:
	//   https://devzone.nordicsemi.com/question/81976/cant-read-or-write-any-twi-registers-after-moving-to-the-app_twi-library/
	/*
	static uint8_t register_addr = KXTJ2_WHO_AM_I;
	static uint8_t register_value = 0x0U;

	// Write before read.
	static app_twi_transfer_t const transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &register_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &register_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	printf("WHO_AM_I: 0x%X\n", register_value);

	if (register_value == 0x0U)
	{
		APP_ERROR_CHECK(BIKE_TRACKER_KIONIX_ERROR_NO_COMMS);
	}
	*/

    m_kionix_state = STATE_INITED;
}

// Send these values to Kionix support if we need help. Cut and paste from RTT. For each read here, we do a write before read.
void acc_kionix_read_all(void)
{
	ret_code_t err_code = 0;

	// WHO_AM_I
	static uint8_t const who_am_i_addr = KXTJ2_WHO_AM_I;
	static uint8_t who_am_i_value = 0x0U;
	static app_twi_transfer_t const who_am_i_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &who_am_i_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &who_am_i_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, who_am_i_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("WHO_AM_I: 0x%X\n", who_am_i_value);

	// WAKEUP_THRESHOLD
	static uint8_t const wake_threshold_addr = KXTJ2_WAKEUP_THRESHOLD;
	static uint8_t wake_threshold_value = 0x0U;
	static app_twi_transfer_t const wake_threshold_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &wake_threshold_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &wake_threshold_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, wake_threshold_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("WAKEUP_THRESHOLD: 0x%X\n", wake_threshold_value);

	// WAKEUP_TIMER
	static uint8_t const wake_timer_addr = KXTJ2_WAKEUP_TIMER;
	static uint8_t wake_timer_value = 0x0U;
	static app_twi_transfer_t const wake_timer_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &wake_timer_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &wake_timer_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, wake_timer_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("WAKEUP_TIMER: 0x%X\n", wake_timer_value);

	// CTRL_REG1
	static uint8_t const ctrl_reg1_addr = KXTJ2_CTRL_REG1;
	static uint8_t ctrl_reg1_value = 0x0U;
	static app_twi_transfer_t const ctrl_reg1_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &ctrl_reg1_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &ctrl_reg1_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, ctrl_reg1_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("CTRL_REG1: 0x%X\n", ctrl_reg1_value);

	// CTRL_REG2
	static uint8_t const ctrl_reg2_addr = KXTJ2_CTRL_REG2;
	static uint8_t ctrl_reg2_value = 0x0U;
	static app_twi_transfer_t const ctrl_reg2_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &ctrl_reg2_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &ctrl_reg2_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, ctrl_reg2_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("CTRL_REG2: 0x%X\n", ctrl_reg2_value);

	// INT_CTRL_REG1
	static uint8_t const int_ctrl_reg1_addr = KXTJ2_INT_CTRL_REG1;
	static uint8_t int_ctrl_reg1_value = 0x0U;
	static app_twi_transfer_t const int_ctrl_reg1_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &int_ctrl_reg1_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &int_ctrl_reg1_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, int_ctrl_reg1_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("INT_CTRL_REG1: 0x%X\n", int_ctrl_reg1_value);

	// INT_CTRL_REG2
	static uint8_t const int_ctrl_reg2_addr = KXTJ2_INT_CTRL_REG2;
	static uint8_t int_ctrl_reg2_value = 0x0U;
	static app_twi_transfer_t const int_ctrl_reg2_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &int_ctrl_reg2_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &int_ctrl_reg2_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, int_ctrl_reg2_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("INT_CTRL_REG2: 0x%X\n", int_ctrl_reg2_value);

	// DATA_CTRL_REG
	static uint8_t const data_ctrl_reg_addr = KXTJ2_DATA_CTRL_REG;
	static uint8_t data_ctrl_reg_value = 0x0U;
	static app_twi_transfer_t const data_ctrl_reg_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &data_ctrl_reg_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &data_ctrl_reg_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, data_ctrl_reg_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("DATA_CTRL_REG: 0x%X\n", data_ctrl_reg_value);

	// STATUS_REG
	static uint8_t const status_reg_addr = KXTJ2_STATUS_REG;
	static uint8_t status_reg_value = 0x0U;
	static app_twi_transfer_t const status_reg_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &status_reg_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &status_reg_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, status_reg_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);
	printf("STATUS_REG: 0x%X\n", status_reg_value);
}

// Program the accelerometer to send us back an interrupt on a given amount of motion.
void acc_kionix_program(kionix_mode_t mode)
{
	ret_code_t err_code = 0;

	if (m_kionix_state >= STATE_PROGRAMMED)
	{
		printf("Already programmed. Nothing to do.\n");
		return;
	}
	else if (m_kionix_state < STATE_INITED)
	{
		APP_ERROR_HANDLER(BIKE_TRACKER_KIONIX_ERROR_INVALID_STATE);
	}

	// Put the accelerometer in stand-by mode by writing 0 to the PC1 bit of control register 1. Or in fact, to all bits (to try
	// interrupt on data for debugging, use 0b00100000).
	static uint8_t const ctrl_reg1_1_tuple[] = { KXTJ2_CTRL_REG1, 0b00000000 };
	static app_twi_transfer_t const ctrl_reg1_1_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, ctrl_reg1_1_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, ctrl_reg1_1_transfers, 1, NULL); // One transfer, no user function.
	APP_ERROR_CHECK(err_code);

	// Output data rate for wake up function (OWUF). Three least significant bits of control register 2. Use 50 Hz.
	//   000: 0.781 Hz
	//   001: 1.563 Hz
	//   010: 3.125 Hz
	//   011: 6.25 Hz
	//   100: 12.5 Hz
	//   101: 25 Hz
	//   110: 50 Hz
	//   111: 100 Hz
	static uint8_t const ctrl_reg2_tuple[] = { KXTJ2_CTRL_REG2, 0b00000110 }; // 0x06
	static app_twi_transfer_t const ctrl_reg2_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, ctrl_reg2_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, ctrl_reg2_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	// Set the minimum amount of acceleration we need for the above duration before firing the interrupt. This is in "counts" where there
	// are 16 counts per g.
	if (mode == KIONIX_MODE_ALERTING)
	{
		static uint8_t const wake_threshold_tuple[] = { KXTJ2_WAKEUP_THRESHOLD, KIONIX_THRESHOLD_FOR_ALERTING };
		static app_twi_transfer_t const wake_threshold_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, wake_threshold_tuple, 2, 0) };
		err_code = app_twi_perform(&m_twi_kionix, wake_threshold_transfers, 1, NULL);
		APP_ERROR_CHECK(err_code);
	}
	else if (mode == KIONIX_MODE_TRACKING)
	{
		static uint8_t const wake_threshold_tuple[] = { KXTJ2_WAKEUP_THRESHOLD, KIONIX_THRESHOLD_FOR_TRACKING };
		static app_twi_transfer_t const wake_threshold_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, wake_threshold_tuple, 2, 0) };
		err_code = app_twi_perform(&m_twi_kionix, wake_threshold_transfers, 1, NULL);
		APP_ERROR_CHECK(err_code);
	}
	else
	{
		APP_ERROR_CHECK(NRF_ERROR_INVALID_PARAM);
	}

	// Set the amount of time we need motion for before firing the interrupt. This is number of clock cycles at the OWUF frequency we set
	// above.
	if (mode == KIONIX_MODE_ALERTING)
	{
		static uint8_t const wake_timer_tuple[] = { KXTJ2_WAKEUP_TIMER, KIONIX_TIME_FOR_ALERTING };
		static app_twi_transfer_t const wake_timer_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, wake_timer_tuple, 2, 0) };
		err_code = app_twi_perform(&m_twi_kionix, wake_timer_transfers, 1, NULL);
		APP_ERROR_CHECK(err_code);
	}
	else if (mode == KIONIX_MODE_TRACKING)
	{
		static uint8_t const wake_timer_tuple[] = { KXTJ2_WAKEUP_TIMER, KIONIX_TIME_FOR_TRACKING };
		static app_twi_transfer_t const wake_timer_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, wake_timer_tuple, 2, 0) };
		err_code = app_twi_perform(&m_twi_kionix, wake_timer_transfers, 1, NULL);
		APP_ERROR_CHECK(err_code);
	}
	else
	{
		APP_ERROR_CHECK(NRF_ERROR_INVALID_PARAM);
	}

	// Set the resolution and range and turn wake on motion mode on. Don't change this without also changing the call where we switch to
	// operating mode at the end.
	//   0... .... PC1. Stand-by mode.
	//   .0.. .... RES. Low current, 8-bit resolution.
	//   ..0. .... DRDYE. Not interested in actual acceleration data, only the interrupt.
	//   ...0 .... GSEL1. +/- 2g range.
	//   .... 0... GSEL0. +/- 2g range.
	//   .... .0.. Reserved
	//   .... ..1. WFUE. Wake up function enable. Send an interrupt on motion.
	//   .... ...0 Reserved
	static uint8_t const ctrl_reg1_2_tuple[] = { KXTJ2_CTRL_REG1, 0b00000010 }; // 0x02
	static app_twi_transfer_t const ctrl_reg1_2_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, ctrl_reg1_2_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, ctrl_reg1_2_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	// Set the interrupt up as active high, latching.
	//   0... .... Reserved
	//   .0.. .... Reserved
	//   ..1. .... IEN. Enable interrupt.
	//   ...1 .... IEA. Active high.
	//   .... 0... IEL. Interrupt pin latches, meaning it will stay high until cleared.
	//   .... .0.. Reserved
	//   .... ..0. STPOL. Self-test polarity.
	//   .... ...0 Reserved
	static uint8_t const int_ctrl_reg1_tuple[] = { KXTJ2_INT_CTRL_REG1, 0b00110000 }; // 0x30
	static app_twi_transfer_t const int_ctrl_reg1_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, int_ctrl_reg1_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, int_ctrl_reg1_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	// Set both directions of all axes as able to cause the interrupt.
	//   0... .... Reserved
	//   .0.. .... Reserved
	//   ..1. .... x -ve
	//   ...1 .... x +ve
	//   .... 1... y -ve
	//   .... .1.. y +ve
	//   .... ..1. z -ve
	//   .... ...1 z +ve
	static uint8_t const int_ctrl_reg2_tuple[] = { KXTJ2_INT_CTRL_REG2, 0b00111111 }; // 0x3F
	static app_twi_transfer_t const int_ctrl_reg2_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, int_ctrl_reg2_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, int_ctrl_reg2_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	// Value as suggested by Kionix. Stream ODR 50 Hz (high res) or 25 Hz (low res).
	static uint8_t const data_ctrl_reg_tuple[] = { KXTJ2_DATA_CTRL_REG, 0b00000010 }; // 0x02
	static app_twi_transfer_t const data_ctrl_reg_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, data_ctrl_reg_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, data_ctrl_reg_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	// Switch to operating mode. This should be the same value as earlier, but with the PC1 bit on.
	//   1... .... PC1. Operating mode.
	//   .0.. .... RES. Low current, 8-bit resolution.
	//   ..0. .... DRDYE. Not interested in actual acceleration data, only the interrupt.
	//   ...0 .... GSEL1. +/1 2g range.
	//   .... 0... GSEL0. +/1 2g range.
	//   .... .0.. Reserved
	//   .... ..1. WFUE. Wake up function enable. Send an interrupt on motion.
	//   .... ...0 Reserved
	static uint8_t const ctrl_reg1_3_tuple[] = { KXTJ2_CTRL_REG1, 0b10000010 }; // 0x82
	static app_twi_transfer_t const ctrl_reg1_3_transfers[] = { APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, ctrl_reg1_3_tuple, 2, 0) };
	err_code = app_twi_perform(&m_twi_kionix, ctrl_reg1_3_transfers, 1, NULL);
	APP_ERROR_CHECK(err_code);

	// Wait for all that to go through, then read everything back.
	// TODO: Remove when tested:
	nrf_delay_ms(100); // Value not tested. 9 transfers at 9 bits each at 2.5 us per bit is 203 us.
	acc_kionix_read_all();

	m_kionix_state = STATE_PROGRAMMED;
}

void acc_kionix_clear_interrupt(void)
{
	ret_code_t err_code = 0;

	if (m_kionix_state < STATE_INITED)
	{
		APP_ERROR_HANDLER(BIKE_TRACKER_KIONIX_ERROR_INVALID_STATE);
	}

	// Read the value of INT_REL (interrupt release) in order to clear the interrupt. The interrupt pin should then go back to inactive
	// (low). We're not interested in the value read back.
	// TODO: Why do we need to read KXTJ2_INT_SOURCE2 as well?

	// INT_SOURCE2:
	static uint8_t const int_source2_addr = KXTJ2_INT_SOURCE2;
	static uint8_t int_source2_value = 0x0U;
	static app_twi_transfer_t const int_source2_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &int_source2_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &int_source2_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, int_source2_transfers, 2, NULL);
	APP_ERROR_CHECK(err_code);

	// INT_REL:
	static uint8_t const int_rel_addr = KXTJ2_INT_REL;
	static uint8_t int_rel_value = 0x0U;
	static app_twi_transfer_t const int_rel_transfers[] =
	{
		APP_TWI_WRITE(KIONIX_SLAVE_ADDRESS, &int_rel_addr, 1, APP_TWI_NO_STOP),
		APP_TWI_READ(KIONIX_SLAVE_ADDRESS, &int_rel_value, 1, 0)
	};

	err_code = app_twi_perform(&m_twi_kionix, int_rel_transfers, 2, NULL);
	APP_ERROR_CHECK(err_code);
}

#endif
